//! # HTU21D Driver
//!
//! `HTU21D` is a driver designed to work with the HTU21D sensors on the i2c protocol.
//! NOTE: This driver has a built in sleep of 50 miliseconds when performing get_temperature and
//! get_humidity operations. To bypass the sleep operation, you can use the write and read method 
//! to implement your own version of the get_temperature and get_humidity operations.
/// 
// --snip--

use {
    embedded_hal::blocking::i2c::{Read, Write, WriteRead},
    std::thread,
    std::time::Duration,
};

/// The available Registries on the THU21D sensor
#[derive(Copy, Clone)]
pub enum Register {
    DefaultAddress = 0x40,
    ReadTempNoHold = 0xF3,
    ReadHumNoHold = 0xF5,
    ReadTempHold = 0xE3,
    ReadHumHold = 0xE5,
    WriteUserRegistry = 0xE6,
    ReadUserRegistry = 0xE7,
    SoftReset = 0xFE,
}

impl Register {
    /// Gives the value of each register defined in the Register enum
    /// Returns the value as  a u8 type
    fn value(self) -> u8 {
        self as u8
    }
}

/// The HTU21D Driver itself, struct contains
/// the i2c connection for write/read operations.
pub struct HTU21D<I2C> {
    /// Each HTU21D must be provided with its own i2c connection
    pub i2c: I2C,
}

impl<I2C, E> HTU21D<I2C>
where
    I2C: WriteRead<Error = E> + Write<Error = E> + Read<Error = E>,
{
    // Returns the i2c connection within this driver once created
    ///
    /// # Arguments
    ///
    /// * `i2c` - A connection to the i2c bus the device is connected to
    ///
    /// # Example
    ///
    /// ```
    /// extern crate HTU21D;
    /// extern crate linux_embedded_hal as hal;
    ///
    /// use HTU21D::HTU21D;
    /// use hal::I2cdev;
    ///
    /// let mut dev = I2cdev::new("/dev/i2c-1").unwrap();
    /// let driver = HTU21D::new(dev);
    ///
    /// ```
    pub fn new(i2c: I2C) -> Result<Self, E> {
        let mut driver = HTU21D { i2c };
        driver.soft_reset()?;
        Ok(driver)
    }

    /// Returns Ok if the soft reset was successful or Err if the write did not complete.
    pub fn soft_reset(&mut self) -> Result<(), E> {
        self.i2c.write(
            Register::DefaultAddress.value(),
            &[Register::SoftReset.value()],
        )?;
        Ok(())
    }
    
    /// Returns the temperature from the sensors
    pub fn get_temperature(&mut self) -> Result<f32, E> {
        let sensor_temp = self.write_register(Register::ReadTempNoHold.value())?;
        let result = -46.85 + (175.72 * sensor_temp);
        Ok(result)
    }

    /// Returns the humidity from the sensor
    pub fn get_humidity(&mut self) -> Result<f32, E> {
        let sensor_humid = self.write_register(Register::ReadHumNoHold.value())?;
        let result = -6.0 + (125.0 * sensor_humid);
        Ok(result)
    }

    /// Performs the write op operation using embedded_hal's write function
    ///
    /// # Arguments
    ///
    /// * `command` - A command to write to the HTU21D's register
    ///
    pub fn write(&mut self, command: u8) -> Result<(), E> {
        self.i2c
            .write(Register::DefaultAddress.value(), &[command])?;

        Ok(())
    }

    /// Performs the read op operation using embedded_hal's read function
    /// NOTE: The returned buffer is the raw buffer without the stripped bytes from the HTU21D device.
    pub fn read(&mut self) -> Result<[u8; 3], E> {
        let mut buffer: [u8; 3] = [0; 3];
        self.i2c
            .read(Register::DefaultAddress.value(), &mut buffer)?;
        Ok(buffer)
    }

    /// Performs the write/read operation using embedded_hal's write/read function
    /// Method performs the write operation,  puts the main thread to sleep for
    /// 50 milisecond and then reads the registries from the HTU21D sensor.
    /// 
    /// Returns the stripped and converted bytes fromt he HTU21D sensor.
    /// As described from the HTU21D sensors documentation:
    /// https://cdn-shop.adafruit.com/datasheets/1899_HTU21D.pdf
    ///
    /// # Arguments
    ///
    /// * `command` - A command to write to the HTU21D's register
    ///
    fn write_register(&mut self, command: u8) -> Result<f32, E> {
        let mut buffer: [u8; 3] = [0; 3];

        self.i2c
            .write(Register::DefaultAddress.value(), &[command])?;

        thread::sleep(Duration::from_millis(50));

        self.i2c
            .read(Register::DefaultAddress.value(), &mut buffer)?;

        let conversion = self.convert_bytes(buffer) as f32 / 65536.0;
        Ok(conversion)
    }

    pub fn convert_bytes(&self, reg: [u8; 3]) -> u16 {
        ((reg[0] as u16) << 8) | reg[1] as u16 & 0xFFFC
    }
}
